/* No user fns here.  Pesch 15apr92. */

/*
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/lock.h>
#include "local.h"


#ifdef WANT_SMALL_STDIO 
int WANT_SMALL_STDIO_findfp_c;
#else

static void
std (ptr, flags, file, data)
     FILE *ptr;
     int flags;
     int file;
     struct _reent *data;
{
  ptr->_p = 0;
  ptr->_r = 0;
  ptr->_w = 0;
  ptr->_flags = flags;
  ptr->_file = file;
  ptr->_bf._base = 0;
  ptr->_bf._size = 0;
  ptr->_lbfsize = 0;
  ptr->_cookie = ptr;

  ptr->_write = __swrite;
  ptr->_read = __sread;
  ptr->_seek = __sseek;
  ptr->_close = __sclose;

#ifndef __SINGLE_THREAD__
  __lock_init_recursive (*(_LOCK_RECURSIVE_T *)&ptr->_lock);
#endif

#ifdef __SCLE
  if (__stextmode(ptr->_file))
    ptr->_flags |= __SCLE;
#endif
}

/*
 * exit() calls _cleanup() through *__cleanup, set whenever we
 * open or buffer a file.  This chicanery is done so that programs
 * that do not use stdio need not link it all in.
 *
 * The name `_cleanup' is, alas, fairly well known outside stdio.
 */



void
_cleanup_r (ptr)
     struct _reent *ptr;
{
  /* (void) _fwalk(fclose); */
  (void) _fwalk (ptr, fflush);	/* `cheating' */
}

#ifndef _REENT_ONLY
void
_cleanup ()
{
  _cleanup_r (_GLOBAL_REENT);
}
#endif


/*
 * __sinit() is called whenever stdio's internal variables must be set up.
 */

void
__sinit (s)
     struct _reent *s;
{
  /* make sure we clean up on exit */
  s->__cleanup = _cleanup_r;	/* conservative */

  s->__sdidinit = 1;

  s->__sglue._next = NULL;
#ifndef _REENT_SMALL
  s->__sglue._niobs = 3;
  s->__sglue._iobs = &s->__sf[0];
#else
  s->__sglue._niobs = 0;
  s->__sglue._iobs = NULL;
  s->_stdin = __sfp(s);
  s->_stdout = __sfp(s);
  s->_stderr = __sfp(s);
#endif

  std (s->_stdin,  __SRD, 0, s);

  /* on platforms that have true file system I/O, we can verify whether stdout 
     is an interactive terminal or not.  For all other platforms, we will
     default to line buffered mode here.  */
#ifdef HAVE_FCNTL
  /* NIOS2: Change to have no buffering on stdout */
  std (s->_stdout, __SWR | __SNBF, 1, s);
#else
  /* NIOS2: Change to have no buffering on stdout */
  std (s->_stdout, __SWR | __SNBF, 1, s);
#endif

  std (s->_stderr, __SWR | __SNBF, 2, s);

}

#endif /* WANT_SMALL_STDIO */
